<?php
// clear; \php\php remove_background.php \art\04 \back\04
// %userprofile%\AppData\Local\Temp
/****************************************************************************/
/* Remove_Background.php - (c) Peardox 2022                                 */
/*                                                                          */
/* A simple script to automate background removal for images using the      */
/* registered version of Deep Art Effects                                   */
/*                                                                          */
/* Usage : remove_background <input_path> <output_path>                     */
/*                                                                          */
/* Sample Usage                                                             */
/*                                                                          */
/* /php/php remove_background.php art noback                                */
/*                                                                          */
/* The above command will take the contents of the directory "art" and      */
/* duplicate it's content in "back" using the same filename but the new     */
/* files will have background removal applied producing "Humans" images     */
/* on a transparent-looking background.                                     */
/*                                                                          */
/* When specifying the output directory you may create multiple levels at   */
/* once by specifying multiple directories - e.g. dir1\dir2\dir3 will make  */
/* the requested directories and output to the full path specified          */
/*                                                                          */
/* The directory names can only include alpha-numeric (a-z, A-Z, 0-9),      */
/* colon (:), stop (.), minus (-) and underscore(_) characters              */
/****************************************************************************/

// Have we got mime_content_type?
if(!function_exists("mime_content_type")) {
	die("Please enable fileinfo in php.ini");
}

// StackOverflow code pinching in action!  
// https://stackoverflow.com/questions/193794/how-can-i-change-a-files-extension-using-php
// Change a file extension (to keep the name part but change the extension e.g. image.jpg => image.png);
function replace_extension($filename, $new_extension) {
    $info = pathinfo($filename);
    return $info['filename'] . '.' . $new_extension;
}

// Does one string begin with another (case insensitive check)
function StartsWith($str, $prefix) {
	if(strlen($prefix) <= strlen($str)) {
		if(strncasecmp($str, $prefix, strlen($prefix)) === 0) {
			return true;
		}
	}
	return false;
}

// Remove any temp directories that start with "tensorflow_native_libraries-"
function remove_dae_temp_files() {
	$dir = sys_get_temp_dir();

	if ($handle = opendir($dir)) {
		while (false !== ($entry = readdir($handle))) {
			if ($entry != "." && $entry != "..") {
				if(StartsWith($entry, "tensorflow_native_libraries-")) {
					if(is_dir($dir . DIRECTORY_SEPARATOR . $entry)) {
						// Junk file probably found
						if(file_exists($dir . DIRECTORY_SEPARATOR . $entry . DIRECTORY_SEPARATOR . "tensorflow_jni.dll")) {
							// Try deleting tensorflow_jni.dll (@ prevents possible error message display)
							@unlink($dir . DIRECTORY_SEPARATOR . $entry . DIRECTORY_SEPARATOR . "tensorflow_jni.dll");
						}
						// Try removing the now hopefully empty dir (@ prevents possible error message display)
						@rmdir($dir . DIRECTORY_SEPARATOR . $entry);
					}
				}
			}
		}
		closedir($handle);
	}
}

// Check pathname is valid
function check_pathname($name) {
	if (!preg_match('/^[\/\w\-\.\_\\\:]+$/', $name)) {
		echo "Path names may only include the characters [A-Z][a-z][0-9] :\\/._ and -\n";
		return false;
	} else {
		return true;
	}
}

// Run the actual DAE command line
function perform_command($infile, $outfile) {
	global $dae;
	
	$cmd = $dae . 
	   " removebackground -input " .
	   $infile . 
	   " -output " . 
	   $outfile;
		   
	if(!shell_exec($cmd . " 2>>log_background.txt")) {
		die("Something went wrong");
	}
}

// Get the time in microseconds
function microtime_float() {
  list($usec, $sec) = explode(" ", microtime());
  return ((float)$usec + (float)$sec);
}

// Allow the script to run for as long as it wants
set_time_limit(0);

// Check DAE is in current directory
if(file_exists("dae.exe")) { // Is the CLI called dae.exe
	$dae = ".\dae.exe"; // I can't be assed typing the full name so I always rename it dae
} else if(file_exists("DeepArtEffectsCli.exe")) {
	$dae = ".\DeepArtEffectsCli.exe";
} else {
	die("Cant find Deep Art Effects CLI\n");
}

// Total runtime
$total_time = 0;

// Start time
$start = microtime_float();

// Initialise the paths
$inpath = "";
$outpath = "";

// Initialise the allowed mimetypes to prevent converting e.g. text files (you can add to this list if you want e.g. image/tiff)
$allowed_mimetypes = array('image/jpeg', 'image/png');

// Fetch the command line parameters
if($argc === 3) {
	$inpath = $argv[1];
	$outpath = $argv[2];
}

// If the command line parameters look bad fail with help
if((check_pathname($inpath) === false) || (check_pathname($outpath) === false) || ($argc !== 3)) {
	// End run with helpful message
	die("remove_background <input_path> <output_path>\n        e.g. remove_background art newart");
}

// Counter for current image
$image_index = 0;

// If the input path doesn't exist then quit
if(!file_exists($inpath) && is_dir($inpath)) {
	die("Can't find directory $inpath\n");
}

// If the output path doesn't exist then create it (can make multiple directory levels all at once)
if(!file_exists($outpath)) {
//	mkdir($outpath, 0700, true);
}

function recurse_dir($indir, $outdir, $scan = true, $relpath = DIRECTORY_SEPARATOR) {
	// We want access to some variables while running the background removal for reporting purposes 
	global $start, $total_time, $image_index, $allowed_mimetypes, $total_files;
	
	// This variable is only meaninful when $scan == true, it's used to count the number of images we're processing
	$filecount = 0;
	$processed_files = 0;
	
	if ($handle = opendir($indir . $relpath)) {
		while (false !== ($entry = readdir($handle))) {
			if ($entry != "." && $entry != "..") {
				// The file we want to remove the background from
				$srcfile = $indir  . $relpath . $entry;
				// Are we looking at a file or a directory?
				if(is_file($srcfile)) {
					// Where we want to save the result (switch extension to png)
					$dstfile = $outdir  . $relpath . replace_extension($entry, "png"); 
					
					if(!file_exists($dstfile)) {
						// Extract the type of file we're about to process
						$mimetype = mime_content_type($srcfile);
						// Is this a valid file type for conversion? i.e. don't try converting a docx etc (not a good idea)
						if(in_array($mimetype, $allowed_mimetypes)) {
							// We run everything twice. The first time we simply count the files we're going to process
							// Knowing how many files we're processing allows us to display progress stats when we actually do stuff
							// The second time through $scan is false so we actually run the DAE CLI commands
							if($scan) {
								$filecount++;
							} else {
								// Run the DAE CLI
								perform_command($srcfile, $dstfile); 
								$processed_files++;

								// Calculate some stats so the user knows how things are going
								$elapsed = microtime_float() - $start;
								$start = microtime_float();
								$total_time += $elapsed;
								$gentime = $total_time / ($image_index + 1);
								$eta = round(($total_files - $image_index) * $gentime);


								// Tell the user what's going on
								echo "Created "; 
								echo $outdir  . $relpath . replace_extension($entry, "png") .
									", time = " . 
									sprintf("%0.3f", $elapsed) . 
									", avg = " . 
									sprintf("%0.3f", $gentime) . 
									" sec, Rate = " . 
									sprintf("%0.3f", 60 / $gentime) . 
									" per min, " .
									sprintf("%d", 3600 / $gentime) . 
									" per hour, ETA " .
									sprintf('%d:%02d:%02d', 
											floor($eta / 3600),
											floor($eta / 60) % 60, 
											$eta % 60) .
									", " . ($total_files - $image_index) . " left\n";
								// Next image...
								$image_index++;
							}							
						}
					}
				} else {
					$filecount += recurse_dir($indir, $outdir, $scan, $relpath . $entry . DIRECTORY_SEPARATOR);
				}
			}
		}
		closedir($handle);
	}
	
	if($processed_files > 0) {
		remove_dae_temp_files();
		// Reset $start to remove tidy up period from timings
		$start = microtime_float();		
	}
	return $filecount;
}

echo "Tidying up old junk files possibly remaining from previous runs\n";
remove_dae_temp_files();

echo "Scanning for files to process\n";
// First scan all the files to get a file count for reporting (quick)
$total_files = recurse_dir($inpath, $outpath);

// There will be a few seconds where nothing appears to be happening so explain what we're about to do
echo "Removing background from $total_files image(s) in $inpath and saving in $outpath\n\n";

// Reset $start
$start = microtime_float();

// Now process all the files (slow)
recurse_dir($inpath, $outpath, false);

// Output the final statistics
$total_time = round($total_time);

echo "\nCreated $image_index images in $outpath which took " .
	sprintf('%d:%02d:%02d', 
		floor($total_time / 3600),
		floor($total_time / 60) % 60, 
		$total_time % 60) . 
	"\n\n";
